@@ -1,10 +1,11 @@
 from configparser import RawConfigParser
 from binascii import hexlify
+from itertools import islice
 import errno
 import os
-import re
 import shutil
 import struct
+import sys
 from zlib import crc32
 
 from .hashindex import NSIndex
@@ -40,14 +41,18 @@ class Repository(object):
     class InvalidRepository(Error):
         """{} is not a valid repository"""
 
+    class CheckNeeded(Error):
+        '''Inconsistency detected. Please run "attic check {}"'''
 
-    def __init__(self, path, create=False):
+    def __init__(self, path, create=False, exclusive=False):
         self.path = path
         self.io = None
         self.lock = None
+        self.index = None
+        self._active_txn = False
         if create:
             self.create(path)
-        self.open(path)
+        self.open(path, exclusive)
 
     def __del__(self):
         self.close()
@@ -71,8 +76,29 @@ class Repository(object):
         with open(os.path.join(path, 'config'), 'w') as fd:
             config.write(fd)
 
-    def open(self, path):
-        self.head = None
+    def get_index_transaction_id(self):
+        indicies = sorted((int(name[6:]) for name in os.listdir(self.path) if name.startswith('index.') and name[6:].isdigit()))
+        if indicies:
+            return indicies[-1]
+        else:
+            return None
+
+    def get_transaction_id(self):
+        index_transaction_id = self.get_index_transaction_id()
+        segments_transaction_id = self.io.get_segments_transaction_id()
+        if index_transaction_id is not None and segments_transaction_id is None:
+            raise self.CheckNeeded(self.path)
+        # Attempt to automatically rebuild index if we crashed between commit
+        # tag write and index save
+        if index_transaction_id != segments_transaction_id:
+            if index_transaction_id is not None and index_transaction_id > segments_transaction_id:
+                replay_from = None
+            else:
+                replay_from = index_transaction_id
+            self.replay_segments(replay_from, segments_transaction_id)
+        return self.get_index_transaction_id()
+
+    def open(self, path, exclusive):
         self.path = path
         if not os.path.isdir(path):
             raise self.DoesNotExist(path)
@@ -80,19 +106,21 @@ class Repository(object):
         self.config.read(os.path.join(self.path, 'config'))
         if not 'repository' in self.config.sections() or self.config.getint('repository', 'version') != 1:
             raise self.InvalidRepository(path)
-        self.lock = UpgradableLock(os.path.join(path, 'config'))
+        self.lock = UpgradableLock(os.path.join(path, 'config'), exclusive)
         self.max_segment_size = self.config.getint('repository', 'max_segment_size')
         self.segments_per_dir = self.config.getint('repository', 'segments_per_dir')
         self.id = unhexlify(self.config.get('repository', 'id').strip())
-        self.rollback()
+        self.io = LoggedIO(self.path, self.max_segment_size, self.segments_per_dir)
 
     def close(self):
         if self.lock:
-            self.rollback()
+            if self.io:
+                self.io.close()
+            self.io = None
             self.lock.release()
             self.lock = None
 
-    def commit(self, rollback=True):
+    def commit(self):
         """Commit transaction
         """
         self.io.write_commit()
@@ -100,24 +128,25 @@ class Repository(object):
         self.write_index()
         self.rollback()
 
-    def _available_indices(self, reverse=False):
-        names = [int(name[6:]) for name in os.listdir(self.path) if re.match('index\.\d+', name)]
-        names.sort(reverse=reverse)
-        return names
+    def get_read_only_index(self, transaction_id):
+        if transaction_id is None:
+            return {}
+        return NSIndex((os.path.join(self.path, 'index.%d') % transaction_id).encode('utf-8'), readonly=True)
 
-    def open_index(self, head, read_only=False):
-        if head is None:
+    def get_index(self, transaction_id, do_cleanup=True):
+        self._active_txn = True
+        self.lock.upgrade()
+        if transaction_id is None:
             self.index = NSIndex.create(os.path.join(self.path, 'index.tmp').encode('utf-8'))
             self.segments = {}
             self.compact = set()
         else:
-            if read_only:
-                self.index = NSIndex((os.path.join(self.path, 'index.%d') % head).encode('utf-8'), readonly=True)
-            else:
-                shutil.copy(os.path.join(self.path, 'index.%d' % head),
-                            os.path.join(self.path, 'index.tmp'))
-                self.index = NSIndex(os.path.join(self.path, 'index.tmp').encode('utf-8'))
-            hints = read_msgpack(os.path.join(self.path, 'hints.%d' % head))
+            if do_cleanup:
+                self.io.cleanup(transaction_id)
+            shutil.copy(os.path.join(self.path, 'index.%d' % transaction_id),
+                        os.path.join(self.path, 'index.tmp'))
+            self.index = NSIndex(os.path.join(self.path, 'index.tmp').encode('utf-8'))
+            hints = read_msgpack(os.path.join(self.path, 'hints.%d' % transaction_id))
             if hints[b'version'] != 1:
                 raise ValueError('Unknown hints file version: %d' % hints['version'])
             self.segments = hints[b'segments']
@@ -127,53 +156,55 @@ class Repository(object):
         hints = {b'version': 1,
                  b'segments': self.segments,
                  b'compact': list(self.compact)}
-        write_msgpack(os.path.join(self.path, 'hints.%d' % self.io.head), hints)
+        transaction_id = self.io.get_segments_transaction_id()
+        write_msgpack(os.path.join(self.path, 'hints.%d' % transaction_id), hints)
         self.index.flush()
         os.rename(os.path.join(self.path, 'index.tmp'),
-                  os.path.join(self.path, 'index.%d' % self.io.head))
+                  os.path.join(self.path, 'index.%d' % transaction_id))
         # Remove old indices
-        current = '.%d' % self.io.head
+        current = '.%d' % transaction_id
         for name in os.listdir(self.path):
             if not name.startswith('index.') and not name.startswith('hints.'):
                 continue
             if name.endswith(current):
                 continue
             os.unlink(os.path.join(self.path, name))
+        self.index = None
 
     def compact_segments(self):
         """Compact sparse segments by copying data into new segments
         """
         if not self.compact:
             return
-
-        def lookup(tag, key):
-            return tag == TAG_PUT and self.index.get(key, (-1, -1))[0] == segment
+        index_transaction_id = self.get_index_transaction_id()
         segments = self.segments
         for segment in sorted(self.compact):
-            if segments[segment] > 0:
-                for tag, key, data in self.io.iter_objects(segment, lookup, include_data=True):
-                    new_segment, offset = self.io.write_put(key, data)
-                    self.index[key] = new_segment, offset
-                    segments.setdefault(new_segment, 0)
-                    segments[new_segment] += 1
-                    segments[segment] -= 1
+            if self.io.segment_exists(segment):
+                for tag, key, offset, data in self.io.iter_objects(segment, include_data=True):
+                    if tag == TAG_PUT and self.index.get(key, (-1, -1)) == (segment, offset):
+                        new_segment, offset = self.io.write_put(key, data)
+                        self.index[key] = new_segment, offset
+                        segments.setdefault(new_segment, 0)
+                        segments[new_segment] += 1
+                        segments[segment] -= 1
+                    elif tag == TAG_DELETE:
+                        if index_transaction_id is None or segment > index_transaction_id:
+                            self.io.write_delete(key)
                 assert segments[segment] == 0
+
         self.io.write_commit()
-        for segment in self.compact:
+        for segment in sorted(self.compact):
             assert self.segments.pop(segment) == 0
             self.io.delete_segment(segment)
         self.compact = set()
 
-    def recover(self, path):
-        """Recover missing index by replaying logs"""
-        start = None
-        available = self._available_indices()
-        if available:
-            start = available[-1]
-        self.open_index(start)
-        for segment, filename in self.io._segment_names():
-            if start is not None and segment <= start:
+    def replay_segments(self, index_transaction_id, segments_transaction_id):
+        self.get_index(index_transaction_id, do_cleanup=False)
+        for segment, filename in self.io.segment_iterator():
+            if index_transaction_id is not None and segment <= index_transaction_id:
                 continue
+            if segment > segments_transaction_id:
+                break
             self.segments[segment] = 0
             for tag, key, offset in self.io.iter_objects(segment):
                 if tag == TAG_PUT:
@@ -190,33 +221,122 @@ class Repository(object):
                         s, _ = self.index.pop(key)
                         self.segments[s] -= 1
                         self.compact.add(s)
-                        self.compact.add(segment)
                     except KeyError:
                         pass
+                    self.compact.add(segment)
+                elif tag == TAG_COMMIT:
+                    continue
+                else:
+                    raise self.CheckNeeded(self.path)
             if self.segments[segment] == 0:
                 self.compact.add(segment)
-        if self.io.head is not None:
+        self.write_index()
+        self.rollback()
+
+    def check(self, repair=False):
+        """Check repository consistency
+
+        This method verifies all segment checksums and makes sure
+        the index is consistent with the data stored in the segments.
+        """
+        error_found = False
+        def report_error(msg):
+            nonlocal error_found
+            error_found = True
+            print(msg, file=sys.stderr)
+
+        assert not self._active_txn
+        try:
+            transaction_id = self.get_transaction_id()
+            current_index = self.get_read_only_index(transaction_id)
+        except Exception:
+            transaction_id = self.io.get_segments_transaction_id()
+            current_index = None
+        if transaction_id is None:
+            transaction_id = self.get_index_transaction_id()
+        if transaction_id is None:
+            transaction_id = self.io.get_latest_segment()
+        if repair:
+            self.io.cleanup(transaction_id)
+        segments_transaction_id = self.io.get_segments_transaction_id()
+        self.get_index(None)
+        for segment, filename in self.io.segment_iterator():
+            if segment > transaction_id:
+                continue
+            try:
+                objects = list(self.io.iter_objects(segment))
+            except (IntegrityError, struct.error):
+                report_error('Error reading segment {}'.format(segment))
+                objects = []
+                if repair:
+                    self.io.recover_segment(segment, filename)
+                    objects = list(self.io.iter_objects(segment))
+            self.segments[segment] = 0
+            for tag, key, offset in objects:
+                if tag == TAG_PUT:
+                    try:
+                        s, _ = self.index[key]
+                        self.compact.add(s)
+                        self.segments[s] -= 1
+                    except KeyError:
+                        pass
+                    self.index[key] = segment, offset
+                    self.segments[segment] += 1
+                elif tag == TAG_DELETE:
+                    try:
+                        s, _ = self.index.pop(key)
+                        self.segments[s] -= 1
+                        self.compact.add(s)
+                    except KeyError:
+                        pass
+                    self.compact.add(segment)
+                elif tag == TAG_COMMIT:
+                    continue
+                else:
+                    report_error('Unexpected tag {} in segment {}'.format(tag, segment))
+        # We might need to add a commit tag if no committed segment is found
+        if repair and segments_transaction_id is None:
+            report_error('Adding commit tag to segment {}'.format(transaction_id))
+            self.io.segment = transaction_id + 1
+            self.io.write_commit()
+            self.io.close_segment()
+        if current_index and not repair:
+            if len(current_index) != len(self.index):
+                report_error('Index object count mismatch. {} != {}'.format(len(current_index), len(self.index)))
+            elif current_index:
+                for key, value in self.index.iteritems():
+                    if current_index.get(key, (-1, -1)) != value:
+                        report_error('Index mismatch for key {}. {} != {}'.format(key, value, current_index.get(key, (-1, -1))))
+        if repair:
+            self.compact_segments()
             self.write_index()
+        else:
+            os.unlink(os.path.join(self.path, 'index.tmp'))
+        self.rollback()
+        return not error_found or repair
 
     def rollback(self):
         """
         """
+        self.index = None
         self._active_txn = False
-        if self.io:
-            self.io.close()
-        self.io = LoggedIO(self.path, self.max_segment_size, self.segments_per_dir)
-        if self.io.head is not None and not os.path.exists(os.path.join(self.path, 'index.%d' % self.io.head)):
-            self.lock.upgrade()
-            self.recover(self.path)
-        self.open_index(self.io.head, read_only=True)
 
-    def _len(self):
+    def __len__(self):
+        if not self.index:
+            self.index = self.get_read_only_index(self.get_transaction_id())
         return len(self.index)
 
-    def get(self, id):
+    def list(self, limit=None, marker=None):
+        if not self.index:
+            self.index = self.get_read_only_index(self.get_transaction_id())
+        return [id_ for id_, _ in islice(self.index.iteritems(marker=marker), limit)]
+
+    def get(self, id_):
+        if not self.index:
+            self.index = self.get_read_only_index(self.get_transaction_id())
         try:
-            segment, offset = self.index[id]
-            return self.io.read(segment, offset, id)
+            segment, offset = self.index[id_]
+            return self.io.read(segment, offset, id_)
         except KeyError:
             raise self.DoesNotExist(self.path)
 
@@ -226,9 +346,7 @@ class Repository(object):
 
     def put(self, id, data, wait=True):
         if not self._active_txn:
-            self._active_txn = True
-            self.lock.upgrade()
-            self.open_index(self.io.head)
+            self.get_index(self.get_transaction_id())
         try:
             segment, _ = self.index[id]
             self.segments[segment] -= 1
@@ -245,21 +363,16 @@ class Repository(object):
 
     def delete(self, id, wait=True):
         if not self._active_txn:
-            self._active_txn = True
-            self.lock.upgrade()
-            self.open_index(self.io.head)
+            self.get_index(self.get_transaction_id())
         try:
             segment, offset = self.index.pop(id)
-            self.segments[segment] -= 1
-            self.compact.add(segment)
-            segment = self.io.write_delete(id)
-            self.compact.add(segment)
-            self.segments.setdefault(segment, 0)
         except KeyError:
             raise self.DoesNotExist(self.path)
-
-    def add_callback(self, cb, data):
-        cb(None, None, data)
+        self.segments[segment] -= 1
+        self.compact.add(segment)
+        segment = self.io.write_delete(id)
+        self.compact.add(segment)
+        self.segments.setdefault(segment, 0)
 
     def preload(self, ids):
         """Preload objects (only applies to remote repositories
@@ -280,16 +393,14 @@ class LoggedIO(object):
     _commit = header_no_crc_fmt.pack(9, TAG_COMMIT)
     COMMIT = crc_fmt.pack(crc32(_commit)) + _commit
 
-    def __init__(self, path, limit, segments_per_dir, capacity=100):
+    def __init__(self, path, limit, segments_per_dir, capacity=90):
         self.path = path
         self.fds = LRUCache(capacity)
-        self.segment = None
+        self.segment = 0
         self.limit = limit
         self.segments_per_dir = segments_per_dir
         self.offset = 0
         self._write_fd = None
-        self.head = None
-        self.cleanup()
 
     def close(self):
         for segment in list(self.fds.keys()):
@@ -297,27 +408,40 @@ class LoggedIO(object):
         self.close_segment()
         self.fds = None  # Just to make sure we're disabled
 
-    def _segment_names(self, reverse=False):
+    def segment_iterator(self, reverse=False):
         for dirpath, dirs, filenames in os.walk(os.path.join(self.path, 'data')):
             dirs.sort(key=int, reverse=reverse)
             filenames = sorted((filename for filename in filenames if filename.isdigit()), key=int, reverse=reverse)
             for filename in filenames:
                 yield int(filename), os.path.join(dirpath, filename)
 
-    def cleanup(self):
+
+    def get_latest_segment(self):
+        for segment, filename in self.segment_iterator(reverse=True):
+            return segment
+        return None
+
+    def get_segments_transaction_id(self):
+        """Verify that the transaction id is consistent with the index transaction id
+        """
+        for segment, filename in self.segment_iterator(reverse=True):
+            if self.is_committed_segment(filename):
+                return segment
+        return None
+
+    def cleanup(self, transaction_id):
         """Delete segment files left by aborted transactions
         """
-        self.head = None
-        self.segment = 0
-        for segment, filename in self._segment_names(reverse=True):
-            if self.is_complete_segment(filename):
-                self.head = segment
-                self.segment = self.head + 1
-                return
-            else:
+        self.segment = transaction_id + 1
+        for segment, filename in self.segment_iterator(reverse=True):
+            if segment > transaction_id:
                 os.unlink(filename)
+            else:
+                break
 
-    def is_complete_segment(self, filename):
+    def is_committed_segment(self, filename):
+        """Check if segment ends with a COMMIT_TAG tag
+        """
         with open(filename, 'rb') as fd:
             try:
                 fd.seek(-self.header_fmt.size, os.SEEK_END)
@@ -358,7 +482,10 @@ class LoggedIO(object):
         except OSError:
             pass
 
-    def iter_objects(self, segment, lookup=None, include_data=False):
+    def segment_exists(self, segment):
+        return os.path.exists(self.segment_filename(segment))
+
+    def iter_objects(self, segment, include_data=False):
         fd = self.get_fd(segment)
         fd.seek(0)
         if fd.read(8) != MAGIC:
@@ -377,16 +504,35 @@ class LoggedIO(object):
             key = None
             if tag in (TAG_PUT, TAG_DELETE):
                 key = rest[:32]
-            if not lookup or lookup(tag, key):
-                if include_data:
-                    yield tag, key, rest[32:]
-                else:
-                    yield tag, key, offset
+            if include_data:
+                yield tag, key, offset, rest[32:]
+            else:
+                yield tag, key, offset
             offset += size
             header = fd.read(self.header_fmt.size)
 
+    def recover_segment(self, segment, filename):
+        self.fds.pop(segment).close()
+        # FIXME: save a copy of the original file
+        with open(filename, 'rb') as fd:
+            data = memoryview(fd.read())
+        os.rename(filename, filename + '.beforerecover')
+        print('attempting to recover ' + filename, file=sys.stderr)
+        with open(filename, 'wb') as fd:
+            fd.write(MAGIC)
+            while len(data) >= self.header_fmt.size:
+                crc, size, tag = self.header_fmt.unpack(data[:self.header_fmt.size])
+                if size < self.header_fmt.size or size > len(data):
+                    data = data[1:]
+                    continue
+                if crc32(data[4:size]) & 0xffffffff != crc:
+                    data = data[1:]
+                    continue
+                fd.write(data[:size])
+                data = data[size:]
+
     def read(self, segment, offset, id):
-        if segment == self.segment:
+        if segment == self.segment and self._write_fd:
             self._write_fd.flush()
         fd = self.get_fd(segment)
         fd.seek(offset)
@@ -424,7 +570,6 @@ class LoggedIO(object):
         header = self.header_no_crc_fmt.pack(self.header_fmt.size, TAG_COMMIT)
         crc = self.crc_fmt.pack(crc32(header) & 0xffffffff)
         fd.write(b''.join((crc, header)))
-        self.head = self.segment
         self.close_segment()
 
     def close_segment(self):
